varying	vec3 l;
varying	vec3 h;
varying vec3 v;
varying vec3 n;
varying	vec3 pp;

uniform sampler3D 	noiseMap;

//
// Soft-tooth func, C1 in (n,n+1) with derivatives breaks in integer points
//

float	smoothToothWave ( in float t )
{
	float	x = fract ( t );

	return 4.0 * x * ( 1.0 - x );
}

//
// Saw-tooth wave.
// Periodic func with values in [0,1] and period of 1
// Derivatives breaks at 0, 0.5, 1, 1.5, 2, ....
//

float	sawWave ( in float t )
{
	t -= floor ( t );

	return t >= 0.5 ? 2.0 * ( 1.0 - t ) : 2.0 * t;
}

//
// Simple C1 periodic func with values in [0,1] and period of 1
//

float	sineWave ( in float t )
{
	return 0.5 * ( 1.0 + sin ( t * 3.1415926 ) );
}

//
// Stripes function
//

vec3	stripes ( in vec3 x, in float f )
{
	vec3	t = 0.5 * ( 1.0 + sin ( f * x * 2.0 * 3.1415926 ) );

	return t * t - 0.5;
}

//
// takes values in [-1,1] range
//

vec3	turbulence ( const in vec3 p, const in float freqScale )
{
	float	sum = 0.0;
	vec3	t   = vec3 ( 0.0 );
	float	f   = 1.0;

	for ( int i = 0; i <= 3; i++ )
	{
		t   += abs ( 2.0 * texture3D ( noiseMap, f * p.xyz ).rgb - vec3 ( 1.0 ) ) / f;
		sum += 1.0 / f;
		f   *= freqScale;
	}

										// remap from [0,sum] to [-1,1]
	return 2.0 * t / sum - vec3 ( 1.0 );
}

//
// Fractal Brownian motion func
//

vec3	fBm ( const in vec3 p, const in float freqScale )
{
	float	sum = 0.0;
	vec3	t   = vec3 ( 0.0 );
	float	f   = 1.0;

	for ( int i = 0; i <= 3; i++ )
	{
		t   += ( 2.0 * texture3D ( noiseMap, f * p.xyz ).rgb - vec3 ( 1.0 ) ) / f;
		sum += 1.0 / f;
		f   *= freqScale;
	}

										// remap from [-sum,sum] to [-1,1]
	return t / sum;
}

//
// Multifractal function
//

vec3	multifractal ( const in vec3 p, const in float offset, const in float freqScale )
{
	float	sum = 0.0;
	vec3	t   = vec3 ( 1.0 );
	float	f   = 1.0;

	for ( int i = 0; i <= 3; i++ )
	{
		t   *= offset + ( 2.0 * texture3D ( noiseMap, f * p.xyz ).rgb - vec3 ( 1.0 ) ) * freqScale;
		f   *= freqScale;
	}

	return t;
}

void main (void)
{
	const vec4	specColor = vec4 ( 1.0 );
	const float	specPower = 30.0;

	vec3	f     = multifractal ( 0.3*pp, 0.6, 2.071371 );
	vec3	color = f;
	vec3	n2    = normalize ( n + f );
	vec3	l2    = normalize ( l );
	vec3	h2    = normalize ( h );
	float	diff  = 0.6 + max ( dot ( n2, l2 ), 0.0 );
	vec4	spec  = specColor * pow ( max ( dot ( n2, h2 ), 0.0 ), specPower );

	gl_FragColor.rgb = diff * color + spec.rgb;
	gl_FragColor.a   = 1.0;
}
