//
// Simple sky fragment shader
//
// Author: Alex V. Boreskoff
//

varying	vec3 pos;
varying	vec3 v;
varying	vec3 l;

uniform	float		cloudDensity;
uniform	float		time;

uniform sampler3D 	noiseMap;

//
// Soft-tooth func, C1 in (n,n+1) with derivatives breaks in integer points
//

float	smoothToothWave ( in float t )
{
	float	x = fract ( t );

	return 4.0 * x * ( 1.0 - x );
}

//
// Saw-tooth wave.
// Periodic func with values in [0,1] and period of 1
// Derivatives breaks at 0, 0.5, 1, 1.5, 2, ....
//

float	sawWave ( in float t )
{
	t -= floor ( t );

	return t >= 0.5 ? 2.0 * ( 1.0 - t ) : 2.0 * t;
}

//
// Simple C1 periodic func with values in [0,1] and period of 1
//

float	sineWave ( in float t )
{
	return 0.5 * ( 1.0 + sin ( t * 3.1415926 ) );
}

//
// Stripes function
//

vec3	stripes ( in vec3 x, in float f )
{
	vec3	t = 0.5 * ( 1.0 + sin ( f * x * 2.0 * 3.1415926 ) );

	return t * t - 0.5;
}

//
// takes values in [-1,1] range
//

vec3	turbulence ( const in vec3 p, const in float freqScale )
{
	float	sum = 0.0;
	vec3	t   = vec3 ( 0.0 );
	float	f   = 1.0;

	for ( int i = 0; i <= 3; i++ )
	{
		t   += abs ( 2.0 * texture3D ( noiseMap, f * p.xyz ).rgb - vec3 ( 1.0 ) ) / f;
		sum += 1.0 / f;
		f   *= freqScale;
	}

										// remap from [0,sum] to [-1,1]
	return 2.0 * t / sum - vec3 ( 1.0 );
}

//
// Fractal Brownian motion func
//

vec3	fBm ( const in vec3 p, const in float freqScale )
{
	float	sum = 0.0;
	vec3	t   = vec3 ( 0.0 );
	float	f   = 1.0;

	for ( int i = 0; i <= 3; i++ )
	{
		t   += ( 2.0 * texture3D ( noiseMap, f * p.xyz ).rgb - vec3 ( 1.0 ) ) / f;
		sum += 1.0 / f;
		f   *= freqScale;
	}

										// remap from [-sum,sum] to [-1,1]
	return t / sum;
}

//
// Multifractal function
//

vec3	multifractal ( const in vec3 p, const in float offset, const in float freqScale )
{
	float	sum = 0.0;
	vec3	t   = vec3 ( 1.0 );
	float	f   = 1.0;

	for ( int i = 0; i <= 3; i++ )
	{
		t   *= offset + ( 2.0 * texture3D ( noiseMap, f * p.xyz ).rgb - vec3 ( 1.0 ) ) * freqScale;
		f   *= freqScale;
	}

	return t;
}

vec4	getSkyColor ( in vec3 v )
{
	const	vec4	skyColorLight  = vec4 ( 0.6, 0.9, 1.0, 1.0 ); // 0.6, 0.9
	const	vec4	skyColorNormal = vec4 ( 0.0, 0.5, 0.6, 1.0 );
	const 	vec3	topDir         = vec3 ( 0.0, 0.0, 1.0 );

										// compute sky color as two gradients -
										// top-> down and sun -> opposite
	float	f1       = max ( 0.0, dot ( v, topDir ) );
	float	f2       = max ( 0.0, dot ( v, normalize ( l ) ) );
	vec4	skyColor = mix ( skyColorNormal, skyColorLight, pow ( f2,  2.0 ) );

	return skyColor;
}

void main ()
{
	const	vec4	cloudColor     = vec4 ( 1.0, 1.0, 1.0, 1.0 );
	const	vec3	maxLight       = vec3 ( 1.0, 0.7, 2.1 );

	vec3	arg  = pos + vec3 ( 0.007*time, 0.006*time, 0.009 * time );
//	vec3	turb = 2*texture3D( noiseMap, arg );
	vec3	turb = 0.5*(turbulence ( arg, 2.137 ) + 1.0);

	vec4	skyColor = getSkyColor ( normalize ( v ) );

										// now compute clouds based on n
	float	factor   = smoothstep ( 0.6, 0.9, turb.x + cloudDensity );
	vec4	cloud    = mix ( skyColor, cloudColor, factor );

	gl_FragColor = cloud;
}

