//
// Rippled water fragment shader
//

uniform samplerCube envMap;
uniform float		time;

varying vec3  n;
varying vec3  e;
varying	vec3  p;

uniform sampler3D 	noiseMap;

//
// Soft-tooth func, C1 in (n,n+1) with derivatives breaks in integer points
//

float	smoothToothWave ( in float t )
{
	float	x = fract ( t );

	return 4.0 * x * ( 1.0 - x );
}

//
// Saw-tooth wave.
// Periodic func with values in [0,1] and period of 1
// Derivatives breaks at 0, 0.5, 1, 1.5, 2, ....
//

float	sawWave ( in float t )
{
	t -= floor ( t );

	return t >= 0.5 ? 2.0 * ( 1.0 - t ) : 2.0 * t;
}

//
// Simple C1 periodic func with values in [0,1] and period of 1
//

float	sineWave ( in float t )
{
	return 0.5 * ( 1.0 + sin ( t * 3.1415926 ) );
}

//
// Stripes function
//

vec3	stripes ( in vec3 x, in float f )
{
	vec3	t = 0.5 * ( 1.0 + sin ( f * x * 2.0 * 3.1415926 ) );

	return t * t - 0.5;
}

//
// takes values in [-1,1] range
//

vec3	turbulence ( const in vec3 p, const in float freqScale )
{
	float	sum = 0.0;
	vec3	t   = vec3 ( 0.0 );
	float	f   = 1.0;

	for ( int i = 0; i <= 3; i++ )
	{
		t   += abs ( 2.0 * texture3D ( noiseMap, f * p.xyz ).rgb - vec3 ( 1.0 ) ) / f;
		sum += 1.0 / f;
		f   *= freqScale;
	}

										// remap from [0,sum] to [-1,1]
	return 2.0 * t / sum - vec3 ( 1.0 );
}

//
// Fractal Brownian motion func
//

vec3	fBm ( const in vec3 p, const in float freqScale )
{
	float	sum = 0.0;
	vec3	t   = vec3 ( 0.0 );
	float	f   = 1.0;

	for ( int i = 0; i <= 3; i++ )
	{
		t   += ( 2.0 * texture3D ( noiseMap, f * p.xyz ).rgb - vec3 ( 1.0 ) ) / f;
		sum += 1.0 / f;
		f   *= freqScale;
	}

										// remap from [-sum,sum] to [-1,1]
	return t / sum;
}

//
// Multifractal function
//

vec3	multifractal ( const in vec3 p, const in float offset, const in float freqScale )
{
	float	sum = 0.0;
	vec3	t   = vec3 ( 1.0 );
	float	f   = 1.0;

	for ( int i = 0; i <= 3; i++ )
	{
		t   *= offset + ( 2.0 * texture3D ( noiseMap, f * p.xyz ).rgb - vec3 ( 1.0 ) ) * freqScale;
		f   *= freqScale;
	}

	return t;
}

void main (void)
{
	const float	eta = 0.9;

    					// Compute reflection vector
	vec3	arg = 0.5*p + vec3 ( 0.04101, -0.0612149, 0.071109 ) * time;
//    vec3	ns = 2 * texture3D ( noiseMap, arg ).xyz - vec3 ( 1 );
	vec3	ns = turbulence ( arg, 2.17 );
  	vec3	en = normalize ( e );
  	vec3	nn = normalize ( n  + 0.3 * ns );

    vec3	r  = reflect   ( en, nn );
    
	r.x = -r.x;

    					// Do a lookup into the environment map.
    vec3 envColor = textureCube ( envMap, r ).rgb;

    					// calc fresnels term.  This allows a view dependant
    					// blend of reflection/refraction
    float fresnel = abs ( dot ( en, nn ) );

    fresnel = clamp ( fresnel, 0.1, 0.9 );

    gl_FragColor = vec4 ( envColor, 1.0 );
}