/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright (c) 2000, D.Palomo van Es                                   *
 *                                                                       *
 * Permission to use, copy, modify, distribute and sell this software    *
 * and its documentation for any purpose is hereby granted without fee,  *
 * provided that the above copyright notice appear in all copies and     *
 * that both that copyright notice and this permission notice appear     *
 * in supporting documentation. This software is provided "as is"        *
 * without express or implied warranty.                                  *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Sample of a First-Person Camera for OpenGL                            *
 * cam.h                                                                 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */		
#include <math.h>
#include "cam.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif


////////////////////////////////////////////////////////////////////////////////
// MatrixMult
////////////////////////////////////////////////////////////////////////////////
void MatrixMult(GLfloat a[3][3],  const GLfloat b[3][3], GLfloat r[3][3]) {
  GLfloat c[3][3];
  int i;
  
  for (i = 0; i < 3; i++) {
    for (int j = 0; j < 3; j++) {
      c[i][j] = 0;
      for (int k = 0; k < 3; k++) {
        c[i][j] += a[i][k] * b[k][j];
      } // for
    } // for
  } // for
  for (i = 0; i < 3; i++) {
    for (int j = 0; j < 3; j++) {
      r[i][j] = c[i][j];
    } // for
  } // for
} // MatrixMult

////////////////////////////////////////////////////////////////////////////////
// MatrixMult//column after
////////////////////////////////////////////////////////////////////////////////
void VectorMatrixMult(GLfloat a[3],  const GLfloat b[3][3], GLfloat c[3]) {
  for (int i = 0; i < 3; i++) {
    c[i] = 0;
    for (int j = 0; j < 3; j++) {
      c[i] += b[i][j] * a[j];
    } // for
  } // for
} // MatrixMult

////////////////////////////////////////////////////////////////////////////////
// Camera::Update_glMatrix
////////////////////////////////////////////////////////////////////////////////
void Camera::Update_glMatrix() {
  glLoadIdentity();
  gluLookAt(eye[0], eye[1], eye[2], look[0], look[1], look[2], up[0] - eye[0], up[1] - eye[1], up[2] - eye[2]);
} // Camera::Update_glMatrix

////////////////////////////////////////////////////////////////////////////////
// Camera::Translate
////////////////////////////////////////////////////////////////////////////////
void Camera::Translate(GLfloat x, GLfloat y, GLfloat z) {
  GLfloat result[3];  // Store the result
  GLfloat temp[3] = {x, y, -z};
  GLfloat matrix[3][3] = {{side[0] - eye[0], up[0] - eye[0], look[0] - eye[0]}, 
  {side[1] - eye[1], up[1] - eye[1], look[1] - eye[1]},
  {side[2] - eye[2], up[2] - eye[2], look[2] - eye[2]}};
  
  VectorMatrixMult(temp, matrix, result);
  for (int i = 0; i < 3; i++) {
    side[i] += result[i];
    up[i] += result[i];
    look[i] += result[i];
    eye[i] += result[i];
  } // for
  Update_glMatrix();
} // Camera::Translate


////////////////////////////////////////////////////////////////////////////////
// Camera::Rotate
////////////////////////////////////////////////////////////////////////////////
void Camera::Rotate(GLfloat angle, axis axe) {
  GLfloat result[3];
  GLfloat l;
  GLfloat rca = angle * 2.0 * M_PI / 360.0f;
  GLfloat iresult[3][3] = {{1.0f, 0.0f, 0.0f},{0.0f, 1.0f, 0.0f}, {0.0f, 0.0f, 1.0f}};
  GLfloat matrix[3][3] = {{1.0f, 0.0f, 0.0f},{0.0f, 1.0f, 0.0f}, {0.0f, 0.0f, 1.0f}};
  GLfloat R_x[3][3] = {{1.0f, 0.0f, 0.0f},  {0.0f, 1.0f, 0.0f}, {0.0f, 0.0f, 1.0f}};
  GLfloat R_y[3][3] = {{1.0f, 0.0f, 0.0f}, {0.0f, 1.0f, 0.0f}, {0.0f, 0.0f, 1.0f}};
  GLfloat v[3], d;
  int i;
  
  // Rotate
  switch (axe) { 
  case X_AXIS:
    v[0] = side[0] - eye[0];                     // create v
    v[1] = side[1] - eye[1];
    v[2] = side[2] - eye[2];
    break;
  case Y_AXIS:
    v[0] = up[0] - eye[0];                     // create v
    v[1] = up[1] - eye[1];
    v[2] = up[2] - eye[2];
    break;
  case Z_AXIS:
    v[0] = look[0] - eye[0];                     // create v
    v[1] = look[1] - eye[1];
    v[2] = look[2] - eye[2];
    break;
  } // switch
  
  matrix[0][0] = cos(rca);// R_z(rca)
  matrix[0][1] = -sin(rca);
  matrix[1][0] = sin(rca);
  matrix[1][1] = cos(rca);
  
  l = sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]); // normalize
  v[0] /= l;
  v[1] /= l;
  v[2] /= l;
  d = sqrt(v[1]*v[1] + v[2]*v[2]);
  if (d != 0) {
    R_x[1][1] = v[2]/d; // R_x(-angle)
    R_x[1][2] = v[1]/d;
    R_x[2][1] = -v[1]/d;
    R_x[2][2] = v[2]/d;
    
    R_y[0][0] = d;   // R_y(-angle)
    R_y[0][2] = v[0];
    R_y[2][0] = -v[0];
    R_y[2][2] = d;
  }
  
  for (i = 0; i < 3; i++) {
    look[i] -= eye[i];
    up[i] -= eye[i];
    side[i] -= eye[i];
  } // for
  
  if (d < -0.001 || d > 0.001) {
    MatrixMult(R_x, R_y, iresult);
    MatrixMult(iresult, matrix, iresult);
    
    R_x[1][2] = -v[1]/d;// R_x(+angle)
    R_x[2][1] = v[1]/d;
    R_y[0][2] = -v[0];  // R_y(+angle)
    R_y[2][0] = v[0];
    
    MatrixMult(iresult, R_y, iresult);
    MatrixMult(iresult, R_x, iresult);
  }
  else {
    switch (axe) {
    case X_AXIS:
      iresult[1][1] = cos(rca);
      iresult[1][2] =-sin(rca);
      iresult[2][1] = sin(rca);
      iresult[2][2] = cos(rca);
      break;
    case Y_AXIS:
      iresult[0][0] = cos(rca);
      iresult[2][0] = sin(rca);
      iresult[0][2] =-sin(rca);
      iresult[2][2] = cos(rca);
      break;
    case Z_AXIS:
      iresult[0][0] = cos(rca);
      iresult[0][1] =-sin(rca);
      iresult[1][0] = sin(rca);
      iresult[1][1] = cos(rca);
      break;
    }
  }
  if (axe != Z_AXIS) {
    VectorMatrixMult(look, iresult, result);
    look[0] = result[0];
    look[1] = result[1];
    look[2] = result[2];
  } // if
  
  if (axe != Y_AXIS) {
    VectorMatrixMult(up, iresult, result);
    up[0] = result[0];
    up[1] = result[1];
    up[2] = result[2];
  } // if
  
  if (axe != X_AXIS) {
    VectorMatrixMult(side, iresult, result);
    side[0] = result[0];
    side[1] = result[1];
    side[2] = result[2];
  } // if
  
  for (i = 0; i < 3; i++) {
    look[i] += eye[i];
    up[i] += eye[i];
    side[i] += eye[i];
  } // for
  Update_glMatrix();
} // Camera::Rotate

////////////////////////////////////////////////////////////////////////////////
// Camera::DebugPrint
////////////////////////////////////////////////////////////////////////////////
void Camera::DebugPrint() {
  printf("eye(%.3f, %.3f, %.3f)look(%.3f, %.3f, %.3f)up(%.3f, %.3f, %.3f)side(%.3f, %.3f, %.3f)\n"
    ,eye[0], eye[1], eye[2],
    look[0], look[1], look[2],
    up[0], up[1], up[2],
    side[0], side[1], side[2]);
} // Camera::DebugPrint

